PYPI_NAMES = {
    "gcsfs": "gcsfs",
    "google.auth": "google-auth",
    "google.oauth2": "google-auth-oauthlib",
    "google.cloud.core": "google-cloud-core",
    "google.cloud.run_v1": "google-cloud-run",
    "google.cloud.run_v2": "google-cloud-run",
    "google.cloud.storage": "google-cloud-storage",
    "google.cloud.logging": "google-cloud-logging",
    "google.cloud.pubsub_v1": "google-cloud-pubsub",
    "google.cloud.bigquery": "google-cloud-bigquery",
    "google.cloud.datastore": "google-cloud-datastore",
    "google.cloud.firestore": "google-cloud-firestore",
    "google.cloud.dataplex_v1": "google-cloud-dataplex",
    "google.cloud.functions_v1": "google-cloud-functions",
    "google.cloud.functions_v2": "google-cloud-functions",
    "google.cloud.scheduler_v1": "google-cloud-scheduler",
    "google.cloud.secretmanager": "google-cloud-secret-manager",
    "google.cloud.resource_manager": "google-cloud-resource-manager",
    "google.cloud.artifactregistry": "google-cloud-artifact-registry",
    "google.cloud.resourcemanager_v3": "google-cloud-resource-manager",
    "google.cloud.artifactregistry_v1": "google-cloud-artifact-registry",
    "yaml": "pyyaml",
    "pandas": "pandas",
    "docker": "docker",
    "pyarrow": "pyarrow",
    "pandas_gbq": "pandas-gbq",
}
DEFAULT_ARGS = {
    "location": "europe-west2",
}
